@@ -16,5 +16,5 @@ dictionary = {
     "⅛": "1/8",
     "⅜": "3/8",
     "⅝": "5/8",
-    "⅞": "7/8"
-}
\ No newline at end of file
+    "⅞": "7/8",
+}
