@@ -1,12 +1,11 @@
 #!/usr/bin/env python
 
-import pandas as pd
 import json
-from .data.meat import dictionary as meat_dictionary
-from .data.ingredients import dictionary as ing_dictionary
+
 import inflect
-from unidecode import unidecode
 import numpy as np
+import pandas as pd
+from unidecode import unidecode
 
 engine = inflect.engine()
 
@@ -22,115 +21,20 @@ def flatten(t):
     return [item for sublist in t for item in sublist]
 
 
-def get_proteins():
-    return list(meat_dictionary.keys())
+import re
 
-
-def get_drinks():
-    return [
-        "cider",
-        "cappuccino",
-        "frappuccino",
-        "coffee",
-        "espresso",
-        "cocktail",
-        "beer",
-        "wine",
-        "rootbeer",
-        "drink",
-        "juice",
-        "tea",
-        "milk",
-        "moccha",
-        "shake",
-        "milkshake",
-        "smoothie",
-        "cocoa",
-        "margarita",
-        "mojito",
-    ]
-
-
-def get_soups():
-    return [
-        "soup",
-        "chowder",
-        "bouillabaisse",
-        "gazpacho",
-        "bisque",
-        "bouillon",
-        "broth",
-        "consomme",
-        "consommé",
-        "velouté",
-        "veloute",
-        "vichyssoise",
-        "skink",
-        "laksa",
-        "ramen",
-        "goulash",
-    ]
-
-
-def get_pasta():
-    return ing_dictionary["pasta"]
-
-
-def get_meats():
-    return ["poultry", "lamb", "game", "pork", "beef", "bacon", "sausage"]
-
-
-def get_seafood():
-    return [
-        "octopus",
-        "clam",
-        "mussel",
-        "scallop",
-        "oyster",
-        "crab",
-        "cockle",
-        "shrimp",
-        "calamari",
-        "crawfish",
-        "fish",
-    ]
-
-
-def get_fatty_fish():
-    return ["salmon", "tuna", "trout", "butterfish", "mackerel"]
-
-
-def get_lean_fish():
-    return [
-        "flounder",
-        "cod",
-        "sole",
-        "red snapper",
-        "bass",
-        "perch",
-        "halibut",
-        "pike",
-        "tilapia",
-    ]
-
-
-#
-# THIS GOES ON THE REAL REAL
-#
-
-from .regex import *
 from .data import vulgar_fractions as vf
-from .data.meat import dictionary as dict_meat
-from .data.ingredients import dictionary as dict_ing
 from .data.constants import (
-    u_substitutions,
     n_substitutions,
     s_substitutions,
-    u_weight_conversions,
-    u_volume_conversions,
     u_conversions,
+    u_substitutions,
+    u_volume_conversions,
+    u_weight_conversions,
 )
-import re
+from .data.ingredients import dictionary as dict_ing
+from .data.meat import dictionary as dict_meat
+from . import regex as rgx
 
 
 def P_vulgar_fractions(phrase):
@@ -144,8 +48,8 @@ def P_parentheses(phrase):
         text = re.sub(r"\([^()]*\)", r"", text)
         return text
 
-    def get_bracket_content(text):
-        return list(re.findall(r"\((.*)\)", text))
+    # def get_bracket_content(text):
+    #     return list(re.findall(r"\((.*)\)", text))
 
     def rm_bracket_content(text):
         return re.sub(r"\([^)]*\)", "", text)
@@ -154,11 +58,15 @@ def P_parentheses(phrase):
 
 
 def P_duplicates(phrase):
-    return re.sub(rf"({UNIT}) \1\b", r"\1", str(phrase))
+    return re.sub(rf"({rgx.UNIT.pattern}) \1\b", r"\1", str(phrase))
 
 
 def P_multi_size_fix(phrase):
-    return re.sub(rf"({Q} {SIZE}) or {Q} {SIZE}", "\1", phrase)
+    return re.sub(
+        rf"({rgx.Q.pattern} {rgx.SIZE.pattern}) or {rgx.Q.pattern} {rgx.SIZE.pattern}",
+        "\1",
+        phrase,
+    )
 
 
 def P_multi_misc_fix(phrase):
@@ -166,12 +74,12 @@ def P_multi_misc_fix(phrase):
 
 
 def P_missing_multiplier_symbol_fix(phrase):
-    pattern = rf"^(?:(?P<multiplier>{Q} )(?P<quantity>{RANGE}|{NUMBER})[- ]?(?P<unit>{UNIT})) (?P<misc>{U_MISC})"
+    pattern = rf"^(?:(?P<multiplier>{rgx.Q.pattern} )(?P<quantity>{rgx.RANGE.pattern}|{rgx.NUMBER.pattern})[- ]?(?P<unit>{rgx.UNIT.pattern})) (?P<misc>{rgx.U_MISC.pattern})"
     return re.sub(pattern, r"\g<multiplier>x \g<quantity> \g<unit>", phrase)
 
 
 def P_quantity_dash_unit_fix(phrase):
-    pattern = rf"(?P<quantity>{RANGE}|{NUMBER})-(?P<unit>{UNIT}) (?P<misc>{U_MISC})"
+    pattern = rf"(?P<quantity>{rgx.RANGE.pattern}|{rgx.NUMBER.pattern})-(?P<unit>{rgx.UNIT.pattern}) (?P<misc>{rgx.U_MISC.pattern})"
     return re.sub(pattern, r"\g<quantity> \g<unit>", phrase)
 
 
@@ -189,7 +97,7 @@ def Q_to_number(val):
         return int(values[0]) / int(values[1])
 
     def range_to_number(string):
-        lower, upper = re.split(R_SEP, string)
+        lower, upper = re.split(rgx.R_SEP.pattern, string)
         return (Q_to_number(lower) + Q_to_number(upper)) / 2
 
     if val != val or not val or val == np.nan:
@@ -198,25 +106,27 @@ def Q_to_number(val):
     val = val.strip()
 
     multiplier = 1
-    if re.match(rf"^({Q}) ?[x\*][ 0-9]", val):
-        match = re.match(rf"^({Q}) ?[x\*](?=[0-9 ])(.*)", val)
+    if re.match(rf"^({rgx.Q.pattern}) ?[x\*][ 0-9]", val):
+        match = re.match(rf"^({rgx.Q.pattern}) ?[x\*](?=[0-9 ])(.*)", val)
         multiplier = Q_to_number(match.group(1))
         val = match.group(2).strip()
 
-    if re.match(rf"^{N_WORD}$", val):
+    if re.match(rf"^{rgx.N_WORD.pattern}$", val):
         val = word_number_to_number(val)
 
-    elif re.match(rf"^{N_WHOLE}$", val) or re.match(rf"^{N_DECIMAL}$", val):
+    elif re.match(rf"^{rgx.N_WHOLE.pattern}$", val) or re.match(
+        rf"^{rgx.N_DECIMAL.pattern}$", val
+    ):
         val = float(val)
 
-    elif re.match(rf"^{RANGE}$", val):
+    elif re.match(rf"^{rgx.RANGE.pattern}$", val):
         val = range_to_number(val)
 
-    elif re.match(rf"^{N_FRACTION}$", val):
+    elif re.match(rf"^{rgx.N_FRACTION.pattern}$", val):
         val = fraction_to_number(val)
 
-    elif re.match(rf"^{N_COMPOSED}$", val):
-        whole_num, fraction = val.split(" ")
+    elif re.match(rf"^{rgx.N_COMPOSED.pattern}$", val):
+        whole_num, fraction = val.split(" ", 1)
         val = float(whole_num) + fraction_to_number(fraction)
 
     try:
@@ -228,35 +138,37 @@ def Q_to_number(val):
 
 
 def Q_unit_split(quantity):
-    quantity, unit = re.search(rf"({Q})?(.*)?", quantity).groups()
+    quantity, unit = re.search(rf"({rgx.Q.pattern})?(.*)?", quantity).groups()
     return pd.Series([quantity, unit])
 
 
 def P_quantity_unit(phrase):
-    quantity, ingredient = re.search(rf"({QUANTITY})?(.+)?", phrase).groups()
-    ingredient = re.sub(rf"^ ?{N_PREP} ", "", ingredient).strip()
+    quantity, ingredient = re.search(
+        rf"({rgx.QUANTITY.pattern})?(.+)?", phrase
+    ).groups()
+    ingredient = re.sub(rf"^ ?{rgx.N_PREP.pattern} ", "", ingredient).strip()
 
-    pods, ingredient = re.match(rf"^({MOD})?(.*)?", ingredient).groups()
+    pods, ingredient = re.match(rf"^({rgx.MOD.pattern})?(.*)?", ingredient).groups()
     ingredient, post_mods = re.match(r"([^,]*)?(?:, (.+))?", ingredient).groups()
 
-    quantity, unit = re.search(rf"({Q})?(.*)?", quantity).groups()
+    quantity, unit = re.search(rf"({rgx.Q.pattern})?(.*)?", quantity).groups()
 
     print(phrase, quantity, unit)
 
-    match = re.search(UNIT, quantity)
+    match = rgx.UNIT.search(quantity)
     unit = match.group() if match else ""
-    unit = re.sub(rf"^{N_PREP} ", "", unit).strip()
-    quantity = re.sub(UNIT, "", quantity).strip()
+    unit = re.sub(rf"^{rgx.N_PREP.pattern} ", "", unit).strip()
+    quantity = re.sub(rgx.UNIT.pattern, "", quantity).strip()
 
-    match = re.match(SIZE, ingredient)
+    match = re.match(rgx.SIZE.pattern, ingredient)
     size = match.group().strip() if match else ""
     for key, values in s_substitutions.items():
         if re.match(rf"(?:{r'|'.join(values)})$", size):
             size = key
             break
 
-    ingredient = re.sub(rf"^{SIZE} ", "", ingredient)
-    quantity = re.sub(rf"^{SIZE} ", "", quantity)
+    ingredient = re.sub(rf"^{rgx.SIZE.pattern} ", "", ingredient)
+    quantity = re.sub(rf"^{rgx.SIZE.pattern} ", "", quantity)
 
     return pd.Series([quantity, unit, size, ingredient])
 
@@ -324,29 +236,36 @@ def I_to_singular(ingredient):
         r"\b"
         + r"|".join(
             [
-                "roma",
-                "kwas",
-                r".+less",
-                r".+\'s",
+                r"^roma$",
+                r"^kwas$",
+                r".+less$",
+                r".+\'s$",
                 r".+us$",
-                "is",
+                r"^is$",
                 r".+ss$",
+                r"molasses$",
+                r"calvados$",
             ]
         )
         + r"\b"
     )
 
+    manual = {r"\bradishes\b": "radish"}
+
+    for key, value in manual.items():
+        ingredient = re.sub(key, value, ingredient)
+
     return " ".join(
         [
             engine.singular_noun(t)
-            if engine.singular_noun(t) and not re.search(exceptions, t)
+            if t and engine.singular_noun(t) and not re.search(exceptions, t)
             else t
             for t in ingredient.split(" ")
         ]
     )
 
 
-def I_to_singular_nlp(ingredient):
+def I_to_singular_nlp(ingredient, nlp):
     return " ".join(
         [t.text if t.tag_ not in ["NNS", "NNPS"] else t.lemma_ for t in nlp(ingredient)]
     )
@@ -373,20 +292,20 @@ def P_juice_zest_fix(phrase):
     ]
 
     phrase = re.sub(
-        rf"(?:^(?:the )?(juice and zest|zest and juice)(?: from| of)?.*?(?P<quantity>{Q})).+(?P<citrus>{r'|'.join(citrus_list)})",
-        "\g<quantity> \g<citrus>",
+        rf"(?:^(?:the )?(juice and zest|zest and juice)(?: from| of)?.*?(?P<quantity>{rgx.Q.pattern})).+(?P<citrus>{r'|'.join(citrus_list)})",
+        "\\g<quantity> \\g<citrus>",
         phrase,
     )
 
     return re.sub(
-        rf"(?:^(?:the )?(?P<part>juice|zest|peel|rind)(?: from| of)?.*?(?P<quantity>{Q})).+(?P<citrus>{r'|'.join(citrus_list)})",
-        "\g<quantity> \g<citrus> \g<part>",
+        rf"(?:^(?:the )?(?P<part>juice|zest|peel|rind)(?: from| of)?.*?(?P<quantity>{rgx.Q.pattern})).+(?P<citrus>{r'|'.join(citrus_list)})",
+        "\\g<quantity> \\g<citrus> \\g<part>",
         phrase,
     )
 
 
 def I_label_protein(ingredient):
-    if not re.search(PROTEIN, ingredient):
+    if not re.search(rgx.PROTEIN.pattern, ingredient):
         return ingredient
 
     for protein, values in dict_meat.items():
@@ -406,27 +325,27 @@ def I_simplify(ingredient):
         return ingredient
 
 
-def ingredient_fixer(ingredient, morequent_ingredients):
-    longest_match = ""
-    most_frequent_match = ""
+# def ingredient_fixer(ingredient, morequent_ingredients):
+#     longest_match = ""
+#     most_frequent_match = ""
 
-    for morequent in morequent_ingredients:
-        # arugula --> arugula leaf
-        if re.search(rf"\b{re.escape(ingredient)}$", morequent):
-            if not len(most_frequent_match):
-                most_frequent_match = morequent
+#     for morequent in morequent_ingredients:
+#         # arugula --> arugula leaf
+#         if re.search(rf"\b{re.escape(ingredient)}$", morequent):
+#             if not len(most_frequent_match):
+#                 most_frequent_match = morequent
 
-        # tasty tomato --> tomato
-        if re.search(rf"\b{re.escape(morequent)}$", ingredient):
-            if len(morequent) > len(longest_match):
-                longest_match = morequent
+#         # tasty tomato --> tomato
+#         if re.search(rf"\b{re.escape(morequent)}$", ingredient):
+#             if len(morequent) > len(longest_match):
+#                 longest_match = morequent
 
-    if longest_match:
-        return longest_match
-    elif most_frequent_match:
-        return most_frequent_match
-    else:
-        return ingredient
+#     if longest_match:
+#         return longest_match
+#     elif most_frequent_match:
+#         return most_frequent_match
+#     else:
+#         return ingredient
 
 
 def plural(label_list):
@@ -467,3 +386,28 @@ def rm_roman_numerals(text):
 
 def rm_accent(text):
     return unidecode(text)
+
+
+def P_multi_color_fix(phrase):
+    COLORS = (
+        rf"(?:{rgx.COLOR.pattern}{rgx.MOD_SEP.pattern})+(?P<last>{rgx.COLOR.pattern})"
+    )
+
+    # each time a COLORS pattern is found, it is replaced by the last color
+    return re.sub(COLORS, r"\g<last>", phrase)
+
+
+def P_multi_adj_fix(phrase):
+    stock_mod_ptrn = r"(fish|vegetable|chicken|beef|pork|lamb|duck|turkey|game|shellfish|crab|lobster|shrimp|clam|oyster|scallop|squid|octopus|cuttlefish)"
+    stock_ptrn = rf"{stock_mod_ptrn} or {stock_mod_ptrn} (stock|broth|bouillon|consomme|cube|juice)"
+
+    return re.sub(stock_ptrn, r"\1 \3", phrase)
+
+
+def P_product_name_fix(phrase):
+    # remove certain product names
+    removable_brands = r"\b(?:s ?& ?w|s ?& ?b|m ?& ?m)\b"
+    phrase = re.sub(removable_brands, " ", phrase)
+    # fix spaces
+    phrase = re.sub(r" +", " ", phrase)
+    return phrase
